/* Copyright © 2002 Red Hat, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Red Hat not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  Red Hat makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * RED HAT DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL RED HAT
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author:  Owen Taylor <otaylor@redhat.com>
 */

#include <glib.h>

#include "composite-funcs.h"

#define FbGet8(v,i)   ((guint16) (guint8) ((v) >> (i)))
#define FbIntMult(a,b,t) ( (t) = (a) * (b) + 0x80, ( ( ( (t)>>8 ) + (t) )>>8 ) )

/*****************************************************************************************/

static const long long mmx_constants[] = {
  0x0000ffff0000LL,     /* 0(c) */
  0x01f0003f001fLL,     /* 8(c) */
  0x008404100840LL,     /* 16(c) */
  0x00f800000000LL,     /* 24(c) */
  0x000000fc0000LL,     /* 32(c) */
  0x0000000000f8LL,     /* 40(c) */
  0x00ff000000000000LL, /* 48(c) */
  0x0000ffff0000ffffLL, /* 56(c) */
  0xffff0000ffff0000LL, /* 64(c) */
  0x000000000000ffffLL  /* 72(c) */
};

#define MMXPrepare()  __asm__ __volatile__ ("pxor %%mm0, %%mm0\n\t"		\
					    "movl $0x80808080, %%eax\n\t"	\
					    "movd %%eax, %%mm1\n\t"		\
					    "punpcklbw %%mm0, %%mm1"		\
					    : /* no output */			\
					    : /* no input */			\
					    : "ax")
  
/* Unpack a 32 bit quantity into a mmx register */
#define Load8888(a,m)		"movd  "a","m"\n\t"        \
				"punpcklbw %%mm0, "m"\n\t"

/* Divide all words of a mmx register by 255 */
#define Div255(m,t)		"paddw %%mm1,"m"\n\t"	\
				"movq "m","t"\n\t"	\
				"psrlw $8,"t"\n\t"	\
				"paddw "t","m"\n\t"	\
				"psrlw $8,"m"\n\t"

#ifdef PSHUFW
#define ExpandA4(a,b)           "pshufw $0xff, "a", "b"\n\t" /*   b: 00Aa00Aa00Aa00Aa */			\
				"pcmpeqw "a", "a"\n\t"								\
				"psrlw $8, "a"\n\t"          /*   a: 00ff00ff00ff00ff */			\
				"pxor "b", "a"\n\t"          /*   a: 255 - Aa|255 - Aa|255 - Aa|255 - Aa */
#else
#define ExpandA4(a,b)           "psrlq $48, "a"\n\t"	   /*   a: 00000000000000Aa */				\
				"movq "a", "b"\n\t"	   /*   b: 00000000000000Aa */			        \
				"psllq $16, "a"\n\t"	   /*   a: 0000000000Aa0000 */			        \
				"por "a", "b"\n\t"	   /*   a: 0000000000Aa00Aa */				\
				"por "b", "a"\n\t"	   /*   b: 0000000000Aa00Aa */				\
				"psllq $32, "a"\n\t"	   /*   a: 00Aa00Aa00000000 */				\
				"por "a", "b"\n\t"	   /*   b: 00Aa00Aa00Aa00Aa */				\
				"pcmpeqw "a", "a"\n\t"	   							\
				"psrlw $8, "a"\n\t"        /*   a: 00ff00ff00ff00ff */				\
				"pxor "b", "a"\n\t"        /*   a: 255 - Aa|255 - Aa|255 - Aa|255 - Aa */
#endif /* PSHUFW */

#define ExpandA4Mult(a,b,c)     ExpandA4(a,b)              /*   a: 255 - Aa|255 - Aa|255 - Aa|255 - Aa */       \
                                                           /*   b: 00Aa00Aa00Aa00Aa */                          \
                                "por 48("c"), "b"\n\t"     /*   b: 00ff00Aa00Aa00Aa */

#ifndef PSHUFW

/* Expand the 32 low bits of mmx register as 00BB00GG00RR
 */
#define Expand8888Rev(m,t1,t2,c) "punpcklbw %%mm0,"m"\n\t"   /*  m: 00Aa00Ba00Ga00Ra */	\
                                 "movq "m","t1"\n\t"					\
                                 "pand 56("c"), "t1"\n\t"    /* t1: 000000Ba000000Ra */	\
                                 "movq "t1","t2"\n\t"	     /* t2: 000000Ba000000Ra */	\
                                 "pand 64("c"), "m"\n\t"     /*  m: 00Aa000000Ga0000 */	\
                                 "psllq $32,"t1"\n\t"	     /* t2: 000000Ra00000000 */	\
                                 "psrlq $32,"t2"\n\t"	     /* t1: 00000000000000Ba */	\
                                 "por "t1","m"\n\t"					\
                                 "por "t2","m"\n\t"          /* m:  00Aa00Ba00Ra00Ga */

#else
/* With pshufw, the above is trivial (and considerably faster) */
#define Expand8888Rev(m,t1,t2,c) "punpcklbw %%mm0,"m"\n\t"	\
                                 "pshufw $0xc6,"m","m"\n\t"
#endif
				
/* Expand the low 16 bits of a mmx register into 00RR00GG00BB

   --- Expanding 565 in the low word ---

    m = (m << (32 - 3)) | (m << (16 - 5)) | m;
    m = m & (01f0003f001f);
    m = m * (008404100840);
    m = m >> 8;
    
    Note the trick here - the top word is shifted by another nibble to avoid
    it bumping into the middle word
  */
#define Expand565(m,t1,t2,c) 	"movq "m", "t1"\n\t"		\
				"movq "m", "t2"\n\t"		\
				"psllq $(36 - 11), "t1"\n\t"	\
				"psllq $(16 - 5), "t2"\n\t"	\
				"por "t1", "m"\n\t"		\
				"por "t2", "m"\n\t"		\
				"pand 8("c"), "m"\n\t"		\
				"pmullw 16("c"), "m"\n\t"       \
                                "psrlw $8, "m"\n\t"

/* Expand the low 16 bits of a mmx register into 00BB00GG00RR */
#define Expand565Rev(m,t1,t2,c) "movq "m", "t1"\n\t"		\
				"movq "m", "t2"\n\t"		\
				"psllq $36, "t1"\n\t"		\
				"psllq $(16 - 5), "t2"\n\t"	\
				"psrlq $11, "m"\n\t"		\
				"por "t1", "m"\n\t"		\
				"por "t2", "m"\n\t"		\
				"pand 8("c"), "m"\n\t"		\
				"pmullw 16("c"), "m"\n\t"	\
                                "psrlw $8, "m"\n\t"
				
/* Pack 00RR00GG00BB into the low 16 bits of a mmx register */
#define Pack565(m,t1,t2,c)	"movq "m", "t1"\n\t"		\
				"movq "m", "t2"\n\t"		\
                                "pand 24("c"), "m"\n\t"		\
                                "pand 32("c"), "t1"\n\t"	\
                                "pand 40("c"), "t2"\n\t"	\
				"psrlq $(32 - 8), "m"\n\t"	\
				"psrlq $(16 - 3), "t1"\n\t"	\
				"psrlq $(0 + 3), "t2"\n\t"	\
 				"por   "t1", "m"\n\t"		\
				"por   "t2", "m"\n\t"

/* Pack 00BB00GG00RR into the low 16 bits of a mmx register */
#define Pack565Rev(m,t1,t2,c)   "movq "m", "t1"\n\t"		\
				"movq "m", "t2"\n\t"		\
                                "pand 24("c"), "m"\n\t"		\
                                "pand 32("c"), "t1"\n\t"	\
                                "pand 40("c"), "t2"\n\t"	\
				"psrlq $(32 + 3), "m"\n\t"	\
				"psrlq $(16 - 3), "t1"\n\t"	\
				"psllq $(8), "t2"\n\t"		\
				"por   "t1", "m"\n\t"		\
				"por   "t2", "m"\n\t"

#define OverARGB(s,d,t1,t2,t3,c)									\
		Load8888(s, t3)              /*  t3: 00Aa00Ra00Ga00Ba */				\
                "movq "t3", "t2"\n\t"	     /*  t2: 00Aa00Ra00Ga00Ba */			        \
                ExpandA4Mult(t3,t1,c)        /*  t1: 00ff00Aa00Aa00Aa */				\
                                             /*  t3: 255 - Aa|255 - Aa|255 - Aa|255 - Aa */		\
		"pmullw "t2", "t1"\n\t"      /*  t1: Aa*255|Aa*Ra|Aa*Ga|Aa*Bb */			\
		"pmullw "t3", "d"\n\t"       /*   d: Aa*(255-Aa)|Rb*(255-Aa)|Gb*(255-Aa)|Bc*(255-Aa) */	\
		Div255(t1,t2)                /*  t1: Aa|Aa*Ra/255|Aa*Ga/255|Aa*sb/255 */		\
		Div255(d,t2)										\
		"paddw "t1", "d"\n\t"        /*   d: 00Ac00Rc00Gc00Bc */
#define OverABGR(s,d,t1,t2,t3,c)									\
		"movd "s", "t3"\t\n"         /*  t3: 00000000AaBaGaRa */				\
                Expand8888Rev(t3,t1,t2,c)    /*  t3: 00Aa00Ra00Ga00Ba */				\
                "movq "t3", "t2"\n\t"	     /*	 t2: 00Aa00Ra00Ga00Ba */				\
                ExpandA4Mult(t3,t1,c)        /*  t1: 00ff00Aa00Aa00Aa */				\
                                             /*  t3: 255 - Aa|255 - Aa|255 - Aa|255 - Aa */		\
		"pmullw "t2", "t1"\n\t"      /*  t1: Aa*255|Aa*Ra|Aa*Ga|Aa*sb */			\
		"pmullw "t3", "d"\n\t"       /*   d: Aa*(255-Aa)|Rb*(255-Aa)|Gb*(255-Aa)|Bc*(255-Aa) */	\
		Div255(t1,t2)                /*  t1: Aa|Aa*Ra/255|Aa*Ga/255|Aa*sb/255 */		\
		Div255(d,t2)										\
		"paddw "t1", "d"\n\t"        /*   d: 00Ac00Rc00Gc00Bc */
#define OverARGBPre(s,d,t1,t2,t3,c)									\
                Load8888(s, t3)              /*  t3: 00Aa00Ra00Ga00Ba */				\
                "movq "t3", "t2"\n\t"	     /*  t2: 00Aa00Ra00Ga00Ba */				\
                ExpandA4(t3, t1)             /*  t3: 255 - Aa|255 - Aa|255 - Aa|255 - Aa */		\
		"pmullw "t3", "d"\n\t"       /*   d: Ab*(255-Aa)|Rb*(255-Aa)|Gb*(255-Aa)|Bc*(255-Aa) */	\
		Div255(d,t1)										\
		"paddw "t2", "d"\n\t"        /*   d: 00Ac00Rc00Gc00Bc */


#define NAME composite_ARGB_0565_pixel
#include "composite-0565-pixel.h"
#undef NAME

#define NAME composite_ABGR_0565_pixel
#define REVERSE
#include "composite-0565-pixel.h"
#undef NAME
#undef REVERSE

#define NAME composite_ARGBpremult_0565_pixel
#define PREMULT
#include "composite-0565-pixel.h"
#undef NAME
#undef PREMULT

#define NAME composite_ARGB_8888_mmx
#define OVER OverARGB
#include "composite-8888-mmx.h"
#undef NAME
#undef OVER

#define NAME composite_ABGR_8888_mmx
#define OVER OverABGR
#include "composite-8888-mmx.h"
#undef NAME
#undef OVER

#define NAME composite_ARGBpremult_8888_mmx
#define OVER OverARGBPre
#include "composite-8888-mmx.h"
#undef NAME
#undef OVER

#define NAME composite_ARGB_0565_mmx
#define PIXEL composite_ARGB_0565_pixel
#define EXPAND Expand565
#define OVER OverARGB
#define PACK Pack565
#include "composite-0565-mmx.h"
#undef NAME
#undef PIXEL
#undef EXPAND
#undef OVER
#undef PACK

#define NAME composite_ABGR_0565_mmx
#define PIXEL composite_ABGR_0565_pixel
#define EXPAND Expand565Rev
#define OVER OverARGB
#define PACK Pack565Rev
#include "composite-0565-mmx.h"
#undef NAME
#undef PIXEL
#undef EXPAND
#undef OVER
#undef PACK

#define NAME composite_ARGBpremult_0565_mmx
#define PIXEL composite_ARGBpremult_0565_pixel
#define EXPAND Expand565
#define OVER OverARGBPre
#define PACK Pack565
#include "composite-0565-mmx.h"
#undef NAME
#undef PIXEL
#undef EXPAND
#undef OVER
#undef PACK

#include "composite-0565-0565-mmx.h"
#include "composite-0565-0565.h"

/************************************************************************************/

#define NAME composite_ARGB_8888
#undef PREMULT
#if G_BYTE_ORDER == G_BIG_ENDIAN
#  define A 0
#  define R 1
#  define G 2
#  define B 3
#else
#  define A 3
#  define R 2
#  define G 1
#  define B 0
#endif
#include "composite-8888.h"
#undef NAME
#undef PREMULT
#undef A
#undef R
#undef G
#undef B

#define NAME composite_ABGR_8888
#undef PREMULT
#if G_BYTE_ORDER == G_BIG_ENDIAN
#  define A 0
#  define B 1
#  define G 2
#  define R 3
#else
#  define A 3
#  define B 2
#  define G 1
#  define R 0
#endif
#include "composite-8888.h"
#undef NAME
#undef PREMULT
#undef A
#undef R
#undef G
#undef B

#define NAME composite_ARGBpremult_8888
#define PREMULT
#if G_BYTE_ORDER == G_BIG_ENDIAN
#  define A 0
#  define R 1
#  define G 2
#  define B 3
#else
#  define A 3
#  define R 2
#  define G 1
#  define B 0
#endif
#include "composite-8888.h"
#undef NAME
#undef PREMULT
#undef A
#undef R
#undef G
#undef B

#define NAME composite_ARGB_0565
#define PIXEL composite_ARGB_0565_pixel
#undef PREMULT
#if G_BYTE_ORDER == G_BIG_ENDIAN
#  define A 0
#  define R 1
#  define G 2
#  define B 3
#else
#  define A 3
#  define R 2
#  define G 1
#  define B 0
#endif
#include "composite-0565.h"
#undef NAME
#undef PIXEL
#undef PREMULT
#undef A
#undef R
#undef G
#undef B

#define NAME composite_ABGR_0565
#define PIXEL composite_ABGR_0565_pixel
#undef PREMULT
#if G_BYTE_ORDER == G_BIG_ENDIAN
#  define A 0
#  define B 1
#  define G 2
#  define R 3
#else
#  define A 3
#  define B 2
#  define G 1
#  define R 0
#endif
#include "composite-0565.h"
#undef NAME
#undef PIXEL
#undef PREMULT
#undef PIXEL
#undef A
#undef R
#undef G
#undef B

#define NAME composite_ARGBpremult_0565
#define PIXEL composite_ARGBpremult_0565_pixel
#define PREMULT
#if G_BYTE_ORDER == G_BIG_ENDIAN
#  define A 0
#  define R 1
#  define G 2
#  define B 3
#else
#  define A 3
#  define R 2
#  define G 1
#  define B 0
#endif
#include "composite-0565.h"
#undef NAME
#undef PIXEL
#undef PREMULT
#undef A
#undef R
#undef G
#undef B
