/* Copyright © 2002 Red Hat, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Red Hat not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  Red Hat makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * RED HAT DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL RED HAT
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author:  Owen Taylor <otaylor@redhat.com>
 */
#include <glib.h>

#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "composite-funcs.h"

#define FbIntMult(a,b,t) ( (t) = (a) * (b) + 0x80, ( ( ( (t)>>8 ) + (t) )>>8 ) )

#define WIDTH 640
#define HEIGHT 400
#define NUM_ITERS 400

typedef void (*ComposeFunc) (guchar *src_buf,
			     gint    src_rowstride,
			     guchar *dest_buf,
			     gint    dest_rowstride,
			     gint    width,
			     gint    height);
typedef void (*ComposeConstantFunc) (guchar *src_buf,
				     gint    src_rowstride,
				     guchar  alpha,
				     guchar *dest_buf,
				     gint    dest_rowstride,
				     gint    width,
				     gint    height);

typedef void (*PrepareFunc) (guchar *src_buf,
			     gint    src_rowstride,
			     gint    width,
			     gint    height);

static void
prepare_ARGBpremult (guchar *src_buf,
		     gint    src_rowstride,
		     gint    width,
		     gint    height)
{
  guchar *src = src_buf;

  while (height--)
    {
      gint twidth = width;
      guchar *p = src;

      while (twidth--)
	{
	  guint t1,t2,t3;

#if G_BYTE_ORDER == G_BIG_ENDIAN
	  p[1] = FbIntMult(p[1],p[0],t1);
	  p[2] = FbIntMult(p[2],p[0],t2);
	  p[3] = FbIntMult(p[3],p[0],t3);
#else
	  p[0] = FbIntMult(p[0],p[3],t1);
	  p[1] = FbIntMult(p[1],p[3],t2);
	  p[2] = FbIntMult(p[2],p[3],t3);
#endif	    

	  p += 4;
	}
      
      src += src_rowstride;
    }
}

static void
prepare_ABGR (guchar *src_buf,
	      gint    src_rowstride,
	      gint    width,
	      gint    height)
{
  guchar *src = src_buf;

  while (height--)
    {
      gint twidth = width;
      guchar *p = src;

      while (twidth--)
	{
	  guchar t;

#if G_BYTE_ORDER == G_BIG_ENDIAN
	  t = p[3];
	  p[3] = p[1];
	  p[1] = t;
#else
	  t = p[2];
	  p[2] = p[0];
	  p[0] = t;
#endif

	  p += 4;
	}
      
      src += src_rowstride;
    }
}

static guchar *
init_buffer (int bpp)
{
  guchar val;
  guchar *buf;
  int j;

  buf = malloc (WIDTH * HEIGHT * bpp * 2);
  
  val = 0;
  for (j = 0; j < WIDTH * HEIGHT * bpp * 2; j++)
    {
      val = (val + ((val + (rand () & 0xff)) >> 1)) >> 1;
      buf[j] = val;
    }

  return buf;
}

/* crc32 code from http://www.cl.cam.ac.uk/Research/SRG/bluebook/21/crc/crc.html,
 * Copyright 1993, Richard Black. This isn't the fastest code in the reference,
 * just the simplest.
 */
static guint crctab[256];
#define QUOTIENT 0x04c11db7

static void
crc32_init(void)
{
  int i, j;
  guint crc;

  for (i = 0; i < 256; i++)
    {
      crc = i << 24;
      for (j = 0; j < 8; j++)
        {
	  if (crc & 0x80000000)
	    crc = (crc << 1) ^ QUOTIENT;
	  else
	    crc = crc << 1;
        }
      crctab[i] = crc;
    }
}

static guint
crc32 (unsigned char *data, int len)
{
  guint result;
  int i;
  static gboolean init = FALSE;

  if (!init)
    {
      crc32_init ();
      init = TRUE;
    }
    
  g_assert (len >= 4);

  result = *data++ << 24;
  result |= *data++ << 16;
  result |= *data++ << 8;
  result |= *data++;
  result = ~ result;
  len -=4;
    
  for (i=0; i<len; i++)
    {
      result = (result << 8 | *data++) ^ crctab[result >> 24];
    }
    
  return ~result;
}

static void
testit (ComposeFunc  func,
	PrepareFunc  prepare,
	int          src_bpp,
	int          dest_bpp,
	char        *name)
{
  GTimer *timer = g_timer_new ();
  gdouble start_time, total_time;
  int i;
  guchar *buf1;
  guchar *buf2;
  
  srand (0);
  buf1 = init_buffer (src_bpp);
  if (prepare)
    prepare (buf1, WIDTH * src_bpp, WIDTH, 2 * HEIGHT);
  
  buf2 = init_buffer (dest_bpp);
 
  start_time = g_timer_elapsed (timer, NULL);
  for (i = 0; i < NUM_ITERS; i++)
    {
      int offset1 = (rand () % (WIDTH * HEIGHT * src_bpp)) & (-src_bpp);
      int offset2 = (rand () % (WIDTH * HEIGHT * dest_bpp)) & (-dest_bpp);
      
      (*func) (buf1 + offset1, src_bpp * WIDTH, buf2 + offset2, dest_bpp * WIDTH, WIDTH, HEIGHT);
    }
  total_time = g_timer_elapsed (timer, NULL) - start_time;

  g_print ("%s time elapsed: %.2fs, %.2f megapixels/s, crc32=%x\n",
	   name,
	   total_time,
	   NUM_ITERS * (WIDTH * HEIGHT * 1e-6) / total_time,
	   crc32 (buf2, 2 * dest_bpp * WIDTH * HEIGHT));

  g_free (buf1);
  g_free (buf2);
  g_timer_destroy (timer);
}

static void
testit_constant (ComposeConstantFunc  func,
		 PrepareFunc  prepare,
		 int          src_bpp,
		 int          dest_bpp,
		 char        *name)
{
  GTimer *timer = g_timer_new ();
  gdouble start_time, total_time;
  int i;
  guchar *buf1;
  guchar *buf2;
  
  srand (0);
  buf1 = init_buffer (src_bpp);
  if (prepare)
    prepare (buf1, WIDTH * src_bpp, WIDTH, 2 * HEIGHT);
  
  buf2 = init_buffer (dest_bpp);
 
  start_time = g_timer_elapsed (timer, NULL);
  for (i = 0; i < NUM_ITERS; i++)
    {
      int offset1 = (rand () % (WIDTH * HEIGHT * src_bpp)) & (-src_bpp);
      int offset2 = (rand () % (WIDTH * HEIGHT * dest_bpp)) & (-dest_bpp);
      guchar alpha = (rand () % 256);

      (*func) (buf1 + offset1, src_bpp * WIDTH, alpha, buf2 + offset2, dest_bpp * WIDTH, WIDTH, HEIGHT);
    }
  total_time = g_timer_elapsed (timer, NULL) - start_time;

  g_print ("%s time elapsed: %.2fs, %.2f megapixels/s, crc32=%x\n",
	   name,
	   total_time,
	   NUM_ITERS * (WIDTH * HEIGHT * 1e-6) / total_time,
	   crc32 (buf2, 2 * dest_bpp * WIDTH * HEIGHT));

  g_free (buf1);
  g_free (buf2);
  g_timer_destroy (timer);
}

int
main (int argc, char **argv)
{
  testit_constant (composite_0565_0565_mmx, NULL, 2, 2, "0565_0565 (mmx)");
  testit_constant (composite_0565_0565, NULL, 2, 2, "0565_0565");
  
  testit (composite_ARGB_8888, NULL, 4, 4, "ARGB_8888");
  testit (composite_ABGR_8888, prepare_ABGR, 4, 4, "ABGR_8888");
  testit (composite_ARGBpremult_8888, prepare_ARGBpremult, 4, 4, "ARGBpremult_8888");

  testit (composite_ARGB_8888_mmx, NULL, 4, 4, "ARGB_8888 (mmx)");
  testit (composite_ABGR_8888_mmx, prepare_ABGR, 4, 4, "ABGR_8888 (mmx)");
  testit (composite_ARGBpremult_8888_mmx, prepare_ARGBpremult, 4, 4, "ARGBpremult_8888 (mmx)");

  testit (composite_ARGB_0565, NULL, 4, 2, "ARGB_0565");
  testit (composite_ABGR_0565, prepare_ABGR, 4, 2, "ABGR_0565");
  testit (composite_ARGBpremult_0565, prepare_ARGBpremult, 4, 2, "ARGBpremult_0565");

  testit (composite_ARGB_0565_mmx, NULL, 4, 2, "ARGB_0565 (mmx)");
  testit (composite_ABGR_0565_mmx, prepare_ABGR, 4, 2, "ABGR_0565 (mmx)");
  testit (composite_ARGBpremult_0565_mmx, prepare_ARGBpremult, 4, 2, "ARGBpremult_0565 (mmx)");

  return 0;
}
